from skyvern.forge import app
from skyvern.forge.sdk.schemas.task_runs import RunEngine, TaskRun, TaskRunResponse, TaskRunType


async def get_task_run(run_id: str, organization_id: str | None = None) -> TaskRun | None:
    return await app.DATABASE.get_task_run(run_id, organization_id=organization_id)


async def get_task_run_response(run_id: str, organization_id: str | None = None) -> TaskRunResponse | None:
    task_run = await get_task_run(run_id, organization_id=organization_id)
    if not task_run:
        return None

    if task_run.task_run_type == TaskRunType.task_v1:
        # fetch task v1 from db and transform to task run response
        task_v1 = await app.DATABASE.get_task(task_run.task_v1_id, organization_id=organization_id)
        if not task_v1:
            return None
        return TaskRunResponse(
            run_id=task_run.run_id,
            engine=RunEngine.skyvern_v1,
            status=task_v1.status,
            goal=task_v1.navigation_goal,
            url=task_v1.url,
            output=task_v1.extracted_information,
            failure_reason=task_v1.failure_reason,
            webhook_url=task_v1.webhook_callback_url,
            totp_identifier=task_v1.totp_identifier,
            totp_url=task_v1.totp_verification_url,
            proxy_location=task_v1.proxy_location,
            created_at=task_v1.created_at,
            modified_at=task_v1.modified_at,
        )
    elif task_run.task_run_type == TaskRunType.task_v2:
        task_v2 = await app.DATABASE.get_task_v2(task_run.task_v2_id, organization_id=organization_id)
        if not task_v2:
            return None
        return TaskRunResponse(
            run_id=task_run.run_id,
            engine=RunEngine.skyvern_v2,
            status=task_v2.status,
            goal=task_v2.prompt,
            url=task_v2.url,
            output=task_v2.output,
            failure_reason=task_v2.failure_reason,
            webhook_url=task_v2.webhook_callback_url,
            totp_identifier=task_v2.totp_identifier,
            totp_url=task_v2.totp_verification_url,
            proxy_location=task_v2.proxy_location,
            created_at=task_v2.created_at,
            modified_at=task_v2.modified_at,
        )
    raise ValueError(f"Invalid task run type: {task_run.task_run_type}")



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from .task_navigation_payload import TaskNavigationPayload
from .proxy_location import ProxyLocation
from .task_extracted_information_schema import TaskExtractedInformationSchema
from .task_type import TaskType
import datetime as dt
from .task_status import TaskStatus
from .task_extracted_information import TaskExtractedInformation
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class Task(UniversalBaseModel):
    title: typing.Optional[str] = pydantic.Field(default=None)
    """
    The title of the task.
    """

    url: str = pydantic.Field()
    """
    Starting URL for the task.
    """

    webhook_callback_url: typing.Optional[str] = pydantic.Field(default=None)
    """
    The URL to call when the task is completed.
    """

    totp_verification_url: typing.Optional[str] = None
    totp_identifier: typing.Optional[str] = None
    navigation_goal: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user's goal for the task.
    """

    data_extraction_goal: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user's goal for data extraction.
    """

    navigation_payload: typing.Optional[TaskNavigationPayload] = pydantic.Field(default=None)
    """
    The user's details needed to achieve the task.
    """

    error_code_mapping: typing.Optional[typing.Dict[str, typing.Optional[str]]] = pydantic.Field(default=None)
    """
    The mapping of error codes and their descriptions.
    """

    proxy_location: typing.Optional[ProxyLocation] = pydantic.Field(default=None)
    """
    The location of the proxy to use for the task.
    """

    extracted_information_schema: typing.Optional[TaskExtractedInformationSchema] = pydantic.Field(default=None)
    """
    The requested schema of the extracted information.
    """

    complete_criterion: typing.Optional[str] = pydantic.Field(default=None)
    """
    Criterion to complete
    """

    terminate_criterion: typing.Optional[str] = pydantic.Field(default=None)
    """
    Criterion to terminate
    """

    task_type: typing.Optional[TaskType] = pydantic.Field(default=None)
    """
    The type of the task
    """

    application: typing.Optional[str] = pydantic.Field(default=None)
    """
    The application for which the task is running
    """

    created_at: dt.datetime = pydantic.Field()
    """
    The creation datetime of the task.
    """

    modified_at: dt.datetime = pydantic.Field()
    """
    The modification datetime of the task.
    """

    task_id: str = pydantic.Field()
    """
    The ID of the task.
    """

    status: TaskStatus = pydantic.Field()
    """
    The status of the task.
    """

    extracted_information: typing.Optional[TaskExtractedInformation] = pydantic.Field(default=None)
    """
    The extracted information from the task.
    """

    failure_reason: typing.Optional[str] = pydantic.Field(default=None)
    """
    The reason for the task failure.
    """

    organization_id: typing.Optional[str] = None
    workflow_run_id: typing.Optional[str] = None
    order: typing.Optional[int] = None
    retry: typing.Optional[int] = None
    max_steps_per_run: typing.Optional[int] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .task_base import TaskBase
from .task_status import TaskStatus
import datetime as dt
import typing
from .task_response_extracted_information import TaskResponseExtractedInformation
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class TaskResponse(UniversalBaseModel):
    request: TaskBase
    task_id: str
    status: TaskStatus
    created_at: dt.datetime
    modified_at: dt.datetime
    extracted_information: typing.Optional[TaskResponseExtractedInformation] = None
    action_screenshot_urls: typing.Optional[typing.List[str]] = None
    screenshot_url: typing.Optional[str] = None
    recording_url: typing.Optional[str] = None
    browser_console_log_url: typing.Optional[str] = None
    downloaded_file_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    max_steps_per_run: typing.Optional[int] = None
    workflow_run_id: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from .task_request_extracted_information_schema import TaskRequestExtractedInformationSchema
from .task_request_navigation_payload import TaskRequestNavigationPayload

__all__ = ["TaskRequestExtractedInformationSchema", "TaskRequestNavigationPayload"]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import pydantic
from .task_base_navigation_payload import TaskBaseNavigationPayload
from .proxy_location import ProxyLocation
from .task_base_extracted_information_schema import TaskBaseExtractedInformationSchema
from .task_type import TaskType
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class TaskBase(UniversalBaseModel):
    title: typing.Optional[str] = pydantic.Field(default=None)
    """
    The title of the task.
    """

    url: str = pydantic.Field()
    """
    Starting URL for the task.
    """

    webhook_callback_url: typing.Optional[str] = pydantic.Field(default=None)
    """
    The URL to call when the task is completed.
    """

    totp_verification_url: typing.Optional[str] = None
    totp_identifier: typing.Optional[str] = None
    navigation_goal: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user's goal for the task.
    """

    data_extraction_goal: typing.Optional[str] = pydantic.Field(default=None)
    """
    The user's goal for data extraction.
    """

    navigation_payload: typing.Optional[TaskBaseNavigationPayload] = pydantic.Field(default=None)
    """
    The user's details needed to achieve the task.
    """

    error_code_mapping: typing.Optional[typing.Dict[str, typing.Optional[str]]] = pydantic.Field(default=None)
    """
    The mapping of error codes and their descriptions.
    """

    proxy_location: typing.Optional[ProxyLocation] = pydantic.Field(default=None)
    """
    The location of the proxy to use for the task.
    """

    extracted_information_schema: typing.Optional[TaskBaseExtractedInformationSchema] = pydantic.Field(default=None)
    """
    The requested schema of the extracted information.
    """

    complete_criterion: typing.Optional[str] = pydantic.Field(default=None)
    """
    Criterion to complete
    """

    terminate_criterion: typing.Optional[str] = pydantic.Field(default=None)
    """
    Criterion to terminate
    """

    task_type: typing.Optional[TaskType] = pydantic.Field(default=None)
    """
    The type of the task
    """

    application: typing.Optional[str] = pydantic.Field(default=None)
    """
    The application for which the task is running
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

EntityType = typing.Union[
    typing.Literal["step", "task", "workflow_run", "workflow_run_block", "observer_thought"], typing.Any
]



from typing import Any, Dict

import structlog
from pydantic import ValidationError

from skyvern.exceptions import UnsupportedActionType
from skyvern.forge.sdk.schemas.tasks import Task
from skyvern.webeye.actions.actions import (
    Action,
    ActionType,
    CheckboxAction,
    ClickAction,
    CompleteAction,
    DownloadFileAction,
    InputTextAction,
    NullAction,
    SelectOption,
    SelectOptionAction,
    SolveCaptchaAction,
    TerminateAction,
    UploadFileAction,
    WaitAction,
)
from skyvern.webeye.scraper.scraper import ScrapedPage

LOG = structlog.get_logger()


def parse_action(action: Dict[str, Any], scraped_page: ScrapedPage, data_extraction_goal: str | None = None) -> Action:
    if "id" in action:
        element_id = action["id"]
    elif "element_id" in action:
        element_id = action["element_id"]
    else:
        element_id = None

    skyvern_element_hash = scraped_page.id_to_element_hash.get(element_id) if element_id else None
    skyvern_element_data = scraped_page.id_to_element_dict.get(element_id) if element_id else None

    reasoning = action["reasoning"] if "reasoning" in action else None
    confidence_float = action["confidence_float"] if "confidence_float" in action else None
    # TODO: currently action intention and response are only used for Q&A actions, like input_text
    # When we start supporting click action, intention will be the reasoning for the click action (why take the action)
    intention = action["user_detail_query"] if "user_detail_query" in action else None
    response = action["user_detail_answer"] if "user_detail_answer" in action else None

    base_action_dict = {
        "element_id": element_id,
        "skyvern_element_hash": skyvern_element_hash,
        "skyvern_element_data": skyvern_element_data,
        "reasoning": reasoning,
        "confidence_float": confidence_float,
        "intention": intention,
        "response": response,
    }

    if "action_type" not in action or action["action_type"] is None:
        return NullAction(**base_action_dict)

    # `.upper()` handles the case where the LLM returns a lowercase action type (e.g. "click" instead of "CLICK")
    action_type = ActionType[action["action_type"].upper()]

    if not action_type.is_web_action():
        # LLM sometimes hallucinates and returns element id for non-web actions such as WAIT, TERMINATE, COMPLETE etc.
        # That can sometimes cause cached action plan to be invalidated. This way we're making sure the element id is not
        # set for non-web actions.
        base_action_dict["element_id"] = None

    if action_type == ActionType.TERMINATE:
        return TerminateAction(**base_action_dict, errors=action["errors"] if "errors" in action else [])

    if action_type == ActionType.CLICK:
        file_url = action["file_url"] if "file_url" in action else None
        return ClickAction(**base_action_dict, file_url=file_url, download=action.get("download", False))

    if action_type == ActionType.INPUT_TEXT:
        return InputTextAction(**base_action_dict, text=action["text"])

    if action_type == ActionType.UPLOAD_FILE:
        # TODO: see if the element is a file input element. if it's not, convert this action into a click action
        return UploadFileAction(
            **base_action_dict,
            file_url=action["file_url"],
        )

    # This action is not used in the current implementation. Click actions are used instead.
    if action_type == ActionType.DOWNLOAD_FILE:
        return DownloadFileAction(**base_action_dict, file_name=action["file_name"])

    if action_type == ActionType.SELECT_OPTION:
        option = action["option"]
        if option is None:
            raise ValueError("SelectOptionAction requires an 'option' field")
        label = option.get("label")
        value = option.get("value")
        index = option.get("index")
        if label is None and value is None and index is None:
            raise ValueError("At least one of 'label', 'value', or 'index' must be provided for a SelectOption")
        return SelectOptionAction(
            **base_action_dict,
            option=SelectOption(
                label=label,
                value=value,
                index=index,
            ),
        )

    if action_type == ActionType.CHECKBOX:
        return CheckboxAction(
            **base_action_dict,
            is_checked=action["is_checked"],
        )

    if action_type == ActionType.WAIT:
        return WaitAction(**base_action_dict)

    if action_type == ActionType.COMPLETE:
        return CompleteAction(
            **base_action_dict,
            data_extraction_goal=data_extraction_goal,
            errors=action["errors"] if "errors" in action else [],
        )

    if action_type == "null":
        return NullAction(**base_action_dict)

    if action_type == ActionType.SOLVE_CAPTCHA:
        return SolveCaptchaAction(**base_action_dict)

    raise UnsupportedActionType(action_type=action_type)


def parse_actions(
    task: Task, step_id: str, step_order: int, scraped_page: ScrapedPage, json_response: list[Dict[str, Any]]
) -> list[Action]:
    actions: list[Action] = []
    for idx, action in enumerate(json_response):
        try:
            action_instance = parse_action(
                action=action, scraped_page=scraped_page, data_extraction_goal=task.data_extraction_goal
            )
            action_instance.organization_id = task.organization_id
            action_instance.workflow_run_id = task.workflow_run_id
            action_instance.task_id = task.task_id
            action_instance.step_id = step_id
            action_instance.step_order = step_order
            action_instance.action_order = idx
            if isinstance(action_instance, TerminateAction):
                LOG.warning(
                    "Agent decided to terminate",
                    task_id=task.task_id,
                    llm_response=json_response,
                    reasoning=action_instance.reasoning,
                    actions=actions,
                )
            actions.append(action_instance)

        except UnsupportedActionType:
            LOG.error(
                "Unsupported action type when parsing actions",
                task_id=task.task_id,
                raw_action=action,
                exc_info=True,
            )
        except (ValidationError, ValueError):
            LOG.warning(
                "Invalid action",
                task_id=task.task_id,
                raw_action=action,
                exc_info=True,
            )
        except Exception:
            LOG.error(
                "Failed to marshal action",
                task_id=task.task_id,
                raw_action=action,
                exc_info=True,
            )

    ############################ This part of code might not be needed ############################
    # Reason #1. validation can be done in action handler but not in parser
    # Reason #2. no need to validate whether the element_id has a hash.
    # If there's no hash, we can fall back to normal operation
    all_element_ids = [action.element_id for action in actions if action.element_id]
    missing_element_ids = [
        element_id for element_id in all_element_ids if element_id not in scraped_page.id_to_element_hash
    ]
    if missing_element_ids:
        LOG.warning(
            "Missing elements in scraped page",
            task_id=task.task_id,
            missing_element_ids=missing_element_ids,
            all_element_ids=all_element_ids,
        )
    ############################ This part of code might not be needed ############################
    return actions



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import datetime as dt
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class TaskGeneration(UniversalBaseModel):
    organization_id: str
    user_prompt: str
    user_prompt_hash: str
    url: typing.Optional[str] = None
    navigation_goal: typing.Optional[str] = None
    navigation_payload: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    data_extraction_goal: typing.Optional[str] = None
    extracted_information_schema: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    llm: typing.Optional[str] = None
    llm_prompt: typing.Optional[str] = None
    llm_response: typing.Optional[str] = None
    suggested_title: typing.Optional[str] = None
    task_generation_id: str
    created_at: dt.datetime
    modified_at: dt.datetime

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .organization_auth_token_type import OrganizationAuthTokenType
import datetime as dt
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import typing
import pydantic


class OrganizationAuthToken(UniversalBaseModel):
    id: str
    organization_id: str
    token_type: OrganizationAuthTokenType = "api"
    token: str
    valid: bool
    created_at: dt.datetime
    modified_at: dt.datetime

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
